from ..py import resolvepy

from b3j0f.utils.ut import UTCase

from unittest import main


class Test(UTCase):

    def test_safe(self):

        self.assertRaises(
            NameError, resolvepy, expr='open', safe=True, besteffort=False
        )

    def test_unsafe(self):

        result = resolvepy(expr='open', safe=False, besteffort=False)

        self.assertIs(open, result)

    def test_tostr(self):

        result = resolvepy(
            expr='open', safe=False, tostr=True, besteffort=False
        )

        self.assertEqual(result, str(open))

    def test_scope(self):

        self.assertRaises(NameError, resolvepy, expr='test', besteffort=False)

        result = resolvepy(expr='test', scope={'test': 1})

        self.assertEqual(result, 1)

    def test_besteffort(self):

        test = 'b3j0f.conf.configurable.log.Logger'

        self.assertRaises(
            NameError, resolvepy, expr=test, besteffort=False
        )

        resolvepy(expr=test, besteffort=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""parser UTs"""

from unittest import main

from b3j0f.utils.ut import UTCase

from ...model.conf import Configuration, configuration
from ...model.cat import category
from ...model.param import Parameter
from ..core import (
    REGEX_REF, REGEX_FORMAT, REGEX_STR, REGEX_EXPR,
    parse, serialize, _ref, ParserError, _strparser
)


class RegexRefTest(UTCase):
    """Test the regex ref."""

    def test_pname(self):

        test = '@test'

        path, cname, history, pname = REGEX_REF.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertFalse(path)
        self.assertFalse(cname)
        self.assertFalse(history)

    def test_history_pname(self):

        test = '@...test'

        path, cname, history, pname = REGEX_REF.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '...')
        self.assertFalse(path)
        self.assertFalse(cname)

    def test_cname_pname(self):

        test = '@cat...test'

        path, cname, history, pname = REGEX_REF.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertEqual(cname, 'cat')
        self.assertFalse(path)

    def test_path_pname(self):

        test = '@ex\@mpl/e/..test'

        path, cname, history, pname = REGEX_REF.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertFalse(cname)
        self.assertEqual(path, 'ex\@mpl/e')

    def test_path_cname_pname(self):

        test = '@ex\@mpl/e/cat...test'

        path, cname, history, pname = REGEX_REF.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertEqual(cname, 'cat')
        self.assertEqual(path, 'ex\@mpl/e')


class RegexFormatTest(UTCase):
    """Test format expression regex."""

    def test_expr(self):

        test = '%test%'

        lang, expr = REGEX_FORMAT.search(test).group('lang', 'expr')

        self.assertFalse(lang)
        self.assertEqual(expr, 'test')

    def test_lang(self):

        test = '%py:test%'

        lang, expr = REGEX_FORMAT.search(test).group('lang', 'expr')

        self.assertEqual(lang, 'py')
        self.assertEqual(expr, 'test')


class RegexStrTest(UTCase):
    """Test str expression regex."""

    def test_expr(self):

        test = '%test%'

        lang, expr = REGEX_STR.search(test).group('lang', 'expr')

        self.assertFalse(lang)
        self.assertEqual(expr, 'test')

    def test_lang(self):

        test = '%py:test%'

        lang, expr = REGEX_STR.search(test).group('lang', 'expr')

        self.assertEqual(lang, 'py')
        self.assertEqual(expr, 'test')

    def test_pname(self):

        test = '@test'

        path, cname, history, pname = REGEX_STR.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertFalse(path)
        self.assertFalse(cname)
        self.assertFalse(history)

    def test_history_pname(self):

        test = '@...test'

        path, cname, history, pname = REGEX_STR.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '...')
        self.assertFalse(path)
        self.assertFalse(cname)

    def test_cname_pname(self):

        test = '@cat...test'

        path, cname, history, pname = REGEX_STR.match(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertEqual(cname, 'cat')
        self.assertFalse(path)

    def test_path_pname(self):

        test = '@ex\@mpl/e/..test'

        path, cname, history, pname = REGEX_STR.search(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertFalse(cname)
        self.assertEqual(path, 'ex\@mpl/e')

    def test_path_cname_pname(self):

        test = '@ex\@mpl/e/cat...test'

        path, cname, history, pname = REGEX_STR.search(test).group(
            'path', 'cname', 'history', 'pname'
        )

        self.assertEqual(pname, 'test')
        self.assertEqual(history, '..')
        self.assertEqual(cname, 'cat')
        self.assertEqual(path, 'ex\@mpl/e')

    def test_both(self):

        test = u'@a%b%@c%js:d%@e'

        matches = REGEX_STR.finditer(test)

        values = [
            ['pname', 'a'],
            ['expr', 'b'],
            ['pname', 'c'],
            ['expr', 'd'],
            ['pname', 'e']
        ]

        for index, match in enumerate(matches):
            groupdict = match.groupdict()
            value = values[index]
            self.assertEqual(groupdict[value[0]], value[1])

    def test_wrong_both(self):

        test = '\%@a%b%\\@k@c%js:d%@e\@\\'

        matches = REGEX_STR.finditer(test)

        values = [
            ['antislash', '%'],
            ['pname', 'a'],
            ['expr', 'b'],
            ['antislash', '@'],
            ['pname', 'c'],
            ['expr', 'd'],
            ['pname', 'e'],
            ['antislash', '@'],
            ['antislash', '\\'],
        ]

        for index, match in enumerate(matches):
            groupdict = match.groupdict()
            value = values[index]
            self.assertEqual(groupdict[value[0]], value[1])


class ExprTest(UTCase):
    """Test full expression regex."""

    def test_expr(self):

        test = '=test'

        lang, expr = REGEX_EXPR.match(test).group('lang', 'expr')

        self.assertFalse(lang)
        self.assertEqual(expr, 'test')

    def test_lang(self):

        test = '=py:test'

        lang, expr = REGEX_EXPR.match(test).group('lang', 'expr')

        self.assertEqual(lang, 'py')
        self.assertEqual(expr, 'test')


class RefTest(UTCase):
    """Test the _ref function."""

    def setUp(self):

        self.pname = 'test'

        self.count = 5

        self.conf = Configuration()

        for i in range(self.count):
            cat = category(str(i), Parameter(name=self.pname, value=i))
            self.conf += cat

    def test_error(self):

        self.assertRaises(ParserError, _ref, pname=self.pname)

    def test_pname(self):

        val = _ref(pname=self.pname, conf=self.conf)

        self.assertEqual(self.count - 1, val.value)

    def test_cname(self):

        val = _ref(pname=self.pname, conf=self.conf, cname=str(self.count - 2))

        self.assertEqual(val.value, self.count - 2)

    def test_history(self):

        val = _ref(pname=self.pname, history=0, conf=self.conf)

        self.assertEqual(val.value, self.count - 1)

        val = _ref(pname=self.pname, history=1, conf=self.conf)

        self.assertEqual(val.value, self.count - 2)

    def test_history_cname(self):

        val = _ref(pname=self.pname, history=0, conf=self.conf, cname=str(self.count - 2))

        self.assertEqual(val.value, self.count - 2)

        val = _ref(pname=self.pname, history=1, conf=self.conf, cname=str(self.count - 2))

        self.assertEqual(val.value, self.count - 3)


class StrParserTest(UTCase):

    def test_empty(self):

        value = _strparser(svalue='')

        self.assertEqual(value, '')

    def test_bool(self):

        val = _strparser(svalue='0', ptype=bool)

        self.assertIs(val, False)

        val = _strparser(svalue='1', ptype=bool)

        self.assertIs(val, True)

        val = _strparser(svalue='true', ptype=bool)

        self.assertIs(val, True)

        val = _strparser(svalue='True', ptype=bool)

        self.assertIs(val, True)

    def test_list(self):

        val = _strparser(svalue='1', ptype=list)

        self.assertEqual(val, ['1'])

        val = _strparser(svalue='', ptype=list)

        self.assertFalse(val)

        val = _strparser(svalue='1, 2, 3', ptype=list)

        self.assertEqual(val, ['1', '2', '3'])

    def test_format(self):

        conf = configuration(category('', Parameter('se', value='es')))

        svalue = 't%"es"%t'

        val = _strparser(svalue=svalue, conf=conf, scope=None)

        self.assertEqual(val, 'test')

    def test_format_expr(self):

        conf = configuration(category('', Parameter('se', value='es')))

        svalue = '%"t"%@se%"t"%'

        val = _strparser(svalue=svalue, conf=conf, scope=None)

        self.assertEqual(val, 'test')

    def test_wrong_format_expr(self):

        conf = configuration(category('', Parameter('se', value='es')))

        svalue = '\@e%"t"%\%@se%"t"%\\'

        val = _strparser(svalue=svalue, conf=conf, scope=None)

        self.assertEqual(val, '@et%est\\')


class ConfigurationTest(UTCase):
    """Base class of test which uses a local configuration."""

    def setUp(self):

        self.count = 5

        self.cnames = [None] * self.count
        self.pname = 'param'
        self.pvalues = [None] * self.count

        self.conf = Configuration()

        for i in range(self.count):
            self.cnames[i] = 'c{0}'.format(i)
            self.pvalues[i] = i + 1
            cat = category(
                self.cnames[i],
                Parameter('param', value=self.pvalues[i])
            )
            self.conf += cat


class SerializerTest(ConfigurationTest):
    """Test the function serializer."""

    def test_str(self):
        """Test to serialize a string."""

        value = 'test'
        serialized = serialize(value)

        self.assertEqual(value, serialized)

    def test_none(self):
        """Test to serialize None."""

        serialized = serialize(None)

        self.assertIsNone(serialized)

    def test_other(self):
        """Test to serialize other."""

        types = [int, float, complex, dict, list, set]

        for _type in types:

            value = _type()

            serialized = serialize(value)

            self.assertEqual(serialized, '=py:{0}'.format(value))


class ParseTest(ConfigurationTest):
    """Test the parse function."""

    def test_default(self):
        """Test default params."""

        value = parse(svalue='=2')

        self.assertEqual(value, 2)

    def test_cname(self):
        """Test with cname."""

        value = parse(svalue='2')

        self.assertEqual(value, '2')

    def test_nocname(self):
        """Test when category name does not exist."""

        value = parse(svalue='t%"es"%t')

        self.assertEqual(value, 'test')

    def test_nopname(self):
        """Test when parameter name does not exist."""

        value = parse(svalue='="test"')

        self.assertEqual(value, 'test')

    def test_expr_ref(self):

        pname = 'test'

        conf = configuration(category('', Parameter('test', value='ify')))

        value = parse(svalue='="test" + @test', conf=conf)

        self.assertEqual(value, 'testify')

    def test_expr_wrong(self):

        pname = 'test'

        conf = configuration(category('', Parameter('test', value='ify')))

        value = parse(svalue='="test" + "\@fgg"', conf=conf)

        self.assertEqual(value, 'test@fgg')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""parser UTs"""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..registry import (
    ResolverRegistry, names, resolve, register, loadresolvers, defaultname,
    getname, unregister
)


class GetNameTest(UTCase):

    def test_routine(self):

        name = getname(lambda **kwargs: None)

        self.assertEqual(name, (lambda: None).__name__)

    def test_class(self):

        name = getname(GetNameTest)

        self.assertEqual(name, GetNameTest.__name__)

    def test_registeredclass(self):

        class Test(object):

            __resolver__ = 'test'

        name = getname(Test)

        self.assertEqual(name, Test.__resolver__)

    def test_callable(self):

        class Test(object):

            def __call__(self):
                pass

        name = getname(Test())

        self.assertEqual(name, Test.__name__)

    def test_registeredcallable(self):

        class Test(object):

            __resolver__ = 'test'

            def __call__(self):
                pass

        name = getname(Test())

        self.assertEqual(name, Test.__resolver__)

    def test_notcallable(self):

        self.assertRaises(TypeError, getname, 2)


class ResolverRegistryTest(UTCase):

    def test_defaultregistry(self):

        reg = ResolverRegistry()

        self.assertIsNone(reg.default)

        self.assertFalse(reg)

    def test_defaultwithregistry(self):

        name = 'test'

        reg = ResolverRegistry(**{name: lambda **_: None})

        self.assertIs(reg.default, name)

        self.assertTrue(reg)

    def test_nodefaultwithoutregistry(self):

        func = lambda **_: None

        reg = ResolverRegistry(default=func)

        self.assertIs(reg.default, func.__name__)

        self.assertTrue(reg)

    def test_defaultandregistry(self):

        reg = ResolverRegistry('test', test=lambda **_: None)

        self.assertEqual(reg.default, 'test')
        self.assertIn('test', reg)

    def test_defaultnotinregistry(self):

        self.assertRaises(
            KeyError, ResolverRegistry, 'test2', test=lambda **_: None
        )

    def test_nonames(self):

        reg = ResolverRegistry()

        self.assertFalse(reg)

    def test_names(self):

        reg = ResolverRegistry(test=lambda **_: None, example=lambda **_: None)

        self.assertEqual(set(reg.keys()), set(['test', 'example']))

    def test_setnotdefault(self):

        reg = ResolverRegistry()

        self.assertRaises(KeyError, setattr, reg, 'default', 'test')

    def test_setdefault(self):

        reg = ResolverRegistry(test=lambda **_: None)

        reg.default = 'test'

        self.assertEqual(reg.default, 'test')

    def test_setnonedefault(self):

        reg = ResolverRegistry(test=lambda **_: None)

        reg.default = None

        self.assertEqual(reg.default, 'test')


class DefaultTest(UTCase):

    def test_name(self):

        _defaultname = defaultname()

        self.assertEqual(names()[0], _defaultname)

    def test_defaultpamplemouse(self):

        pamplemouse = 'pamplemouse'

        self.assertRaises(KeyError, defaultname, name=pamplemouse)

        register(name=pamplemouse, exprresolver=lambda x: pamplemouse)

        _defaultname = defaultname(pamplemouse)

        self.assertEqual(_defaultname, pamplemouse)

        unregister(name=pamplemouse)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Parameter definition objects."""

from __future__ import absolute_import

__all__ = ['Parameter', 'PType', 'BOOL', 'Array', 'ARRAY']

from .base import ModelElement
from ..parser.core import parse, serialize

from parser import ParserError

from six import string_types, reraise

from re import compile as re_compile

from collections import Iterable

from b3j0f.utils.path import lookup

from ..parser.resolver.core import (
    DEFAULT_SAFE, DEFAULT_BESTEFFORT, DEFAULT_SCOPE
)


class PType(object):
    """Dedicated to embed a specific type such as a parameter type in order to
    instanciate easily this embedded type from a parameter serialized value.

    Instanciation depends on type of serialiazed value (by order):

    - str: first arg to this ptype.
    - dict: it is given such as a kwargs to this ptype.
    - iterable: it is given such as an args to this ptype.
    - object: it is given such as the only one argument to this ptype.
    """

    __slots__ = ('ptype')

    def __init__(self, ptype, *args, **kwargs):

        super(PType, self).__init__(*args, **kwargs)

        self.ptype = ptype

    def __instancecheck__(self, instance):
        """Check instance such as this instance or self ptype instance."""

        return isinstance(instance, (PType, self.ptype))

    def __subclasscheck__(self, subclass):
        """Check subclass such as this subclass or self ptype subclass."""

        return issubclass(subclass, (PType, self.ptype))

    def __call__(self, svalue):
        """Instantiate a new instance of this ptype related to input value.

        :param ssvalue: Instanciation depends on type of serialiazed value
            (by order):

        - dict: it is given such as a kwargs to this ptype.
        - iterable: it is given such as an args to this ptype.
        - object: it is given such as the only one argument to this ptype."""

        result = None

        args, kwargs = [], {}

        if isinstance(svalue, string_types) or not isinstance(svalue, Iterable):
            args = [svalue]

        elif isinstance(svalue, dict):
            kwargs = svalue

        else:
            args = svalue

        try:
            result = self.ptype(*args, **kwargs)

        except TypeError:
            raise ParserError(
                'Wrong parameters {0} {1} for {2}'.format(
                    args, kwargs, self.ptype
                )
            )

        return result


class _Bool(PType):
    """Parameter type dedicated to boolean values."""

    def __init__(self, *args, **kwargs):

        super(_Bool, self).__init__(ptype=bool, *args, **kwargs)

    def __call__(self, svalue):

        return svalue in ('true', 'True', '1')


BOOL = _Bool()


class Array(PType):
    """Parameter type dedicated to array value with item type.

    For example, in using the Array(int) permits to convert the entry "2,3,4" to
    [2,3,4] where items are integers."""

    def __init__(self, ptype=object, *args, **kwargs):

        super(Array, self).__init__(ptype=ptype, *args, **kwargs)

    def __instancecheck__(self, instance):
        """Check instance such as this instance or self ptype instance."""

        result = isinstance(instance, Iterable) and not isinstance(
            instance, string_types
        )

        if result:
            for item in instance:
                result = isinstance(item, self.ptype)

                if not result:
                    break

        return result

    def __subclasscheck__(self, subclass):
        """Check subclass such as this subclass or self ptype subclass."""

        return issubclass(subclass, Iterable) and not issubclass(
            subclass, string_types
        )

    def __call__(self, svalue):

        result = []

        if svalue:

            for item in svalue.split(','):

                item = item.strip()

                if not isinstance(item, self.ptype):

                    try:
                        item = self.ptype(item)

                    except (TypeError, ValueError):

                        item = lookup(item)

                        if issubclass(item, self.ptype):

                            item = item()

                if isinstance(item, self.ptype):

                    result.append(item)

                else:
                    raise TypeError(
                        'Wrong item type, {0} expected'.format(self.ptype)
                    )

        return result


ARRAY = Array()  # default array


class Parameter(ModelElement):
    """Parameter identified among a category by its name.

    Provide a value (None by default) and a parser (str by default).

    Properties are:

    - name: parameter name. If regex, drivers will generated as many parameters
        as they exist in the configuration resource with this properties (
        value, conf, parser, etc.).
    - value: parameter value. None if error is not None.
    - error: parameter error encountered while attempting to set values.
    - parser: parameter value parser from configuration resources.
    - conf: parameter configuration if not None. If not None, the value must be
        a callable beceause the conf is used such as a kwargs.
    - local: distinguish local parameters from those found in configuration
        resources.
    """

    __slots__ = (
        '_name', 'ptype', 'parser', '_svalue', '_value', '_error', 'conf',
        'local', 'scope', 'configurable', 'serializer', 'besteffort', 'safe'
    ) + ModelElement.__slots__

    class Error(Exception):
        """Handle Parameter errors."""

    CONF_SUFFIX = '::conf'  #: parameter configuration name.

    PARAM_NAME_REGEX = r'[a-zA-Z_]\w*'  #: simple name validation.
    #: simple name regex compiler.
    _PARAM_NAME_COMPILER_MATCHER = re_compile(PARAM_NAME_REGEX).match

    DEFAULT_NAME = re_compile('.*')
    DEFAULT_PTYPE = None  #: default ptype.
    DEFAULT_LOCAL = True  #: default local value.
    DEFAULT_ERROR = None  #: default error value

    def __init__(
            self, name=DEFAULT_NAME, value=None, ptype=DEFAULT_PTYPE,
            parser=parse, serializer=serialize, svalue=None,
            conf=None, configurable=None,
            local=DEFAULT_LOCAL, scope=DEFAULT_SCOPE, safe=DEFAULT_SAFE,
            besteffort=DEFAULT_BESTEFFORT, error=None,
            *args, **kwargs
    ):
        """
        :param name: depends on type:
            - str: unique by category.
            - regex: designates a group of parameters with a matching name and
                common properties (parser, value, conf, etc.).

            Default is the regex ``.*``.
        :param value: param value. None if not given.
        :param type ptype: parameter value type.
        :param callable parser: param value deserializer which takes in param a
            str. Default is the expression parser.
        :param callable serializer: param serializer which takes in param an
            object and retuns a string. Default is the expression serializer.
        :param str svalue: serialized value.
        :param dict conf: parameter configuration used to init the parameter
            value if not None. In this case, the parameter value must be
            callable.
        :param Configurable configurable: specific configuable.
        :param bool local: distinguish local parameters from those defined
            outside the configuration while updating the embedding conf.
        :param dict scope: dictionary of elements to add in the serialized
            value evaluation.
        """

        super(Parameter, self).__init__(*args, **kwargs)

        # init protected attributes
        self._name = None
        self._value = value
        self._error = error
        self._svalue = svalue

        # init public attributes
        self.parser = parser
        self.serializer = serializer
        self.ptype = ptype
        self.conf = conf
        self.configurable = configurable
        self.local = local
        self.scope = scope
        self.besteffort = besteffort
        self.safe = safe

        # init setters
        self.name = name
        #self.svalue = svalue
        #self.value = value

    def __eq__(self, other):
        """
        :return: True iif other is a parameter and has the same name than
            this.
        :rtype: bool"""

        result = isinstance(other, Parameter)

        if result:
            if isinstance(self.name, string_types):
                if isinstance(other.name, string_types):
                    result = self.name == other.name

                else:
                    result = other.name.match(self.name)

            else:
                if isinstance(other.name, string_types):
                    result = self.name.match(other.name)

                else:
                    result = self.name.pattern == other.name.pattern

        return result

    def __hash__(self):
        """Get parameter hash value.

        :return:hash(self.name)+hash(Parameter).
        :rtype: int"""

        return hash(self.name) + hash(Parameter)

    @property
    def error(self):
        """Get encountered error.

        :rtype: Exception"""

        return self._error

    @property
    def name(self):
        """Get parameter name.

        :return: this name.
        :rtype: str
        """

        return self._name

    @name.setter
    def name(self, value):
        """Set parameter name.

        :param str value: name value.
        """

        if isinstance(value, string_types):

            match = Parameter._PARAM_NAME_COMPILER_MATCHER(value)

            if match is None or match.group() != value:
                value = re_compile(value)

        self._name = value

    @property
    def conf_name(self):
        """Get this configuration name.

        :return: {self.name}{CONF_SUFFIX}.
        :rtype: str
        """

        return '{0}{1}'.format(self.name, Parameter.CONF_SUFFIX)

    @property
    def svalue(self):
        """Get serialized value.

        :rtype: str
        """

        result = self._svalue

        if result is None:  # try to get svalue from value if svalue is None

            try:
                value = self.value

            except Parameter.Error:
                pass

            else:
                result = self._svalue = self.serializer(value)

        return result

    @svalue.setter
    def svalue(self, value):
        """Change of serialized value.

        Nonify this value as well.

        :param str value: serialized value to use.
        """

        if value is not None:  # if value is not None
            self._value = None
            self._error = None

        self._svalue = value  # set svalue

    def resolve(
            self,
            configurable=None, conf=None, scope=None, ptype=None,
            parser=None, error=True, svalue=None,
            safe=None, besteffort=None
    ):
        """Resolve this parameter value related to a configurable and a
        configuration.

        Save error in this attribute `error` in case of failure.

        :param str svalue: serialized value too resolve. Default is this svalue.
        :param Configurable configurable: configurable to use for foreign
            parameter resolution.
        :param Configuration conf: configuration to use for
            cross-value resolution.
        :param dict scope: variables to use for local expression evaluation.
        :param type ptype: return type. Default is this ptype.
        :param parser: specific parser to use. Default this parser.
        :param bool error: raise an error if True (False by default).
        :param bool safe: if True (default) resolve without builtins functions.
        :param bool besteffort: best effort flag. Default is this besteffort.
        :return: newly resolved value.
        :raises: Parameter.Error for any raised exception.
        """

        result = self._value

        # if cached value is None and serialiazed value exists
        if self._value is None and self._svalue is not None:

            self._error = None  # nonify error.

            if ptype is None:
                ptype = self.ptype

            if parser is None:  # init parser
                parser = self.parser

            if svalue is None:
                svalue = self._svalue

            if conf is None:  # init conf
                conf = self.conf

            if configurable is None:  # init configurable
                configurable = self.configurable

            if scope is None:
                scope = self.scope

            else:
                scope, selfscope = self.scope.copy(), scope
                scope.update(selfscope)

            if safe is None:
                safe = self.safe

            if besteffort is None:
                besteffort = self.besteffort

            # parse value if str and if parser exists
            try:
                result = self._value = parser(
                    svalue=svalue, conf=conf,
                    configurable=configurable, ptype=ptype,
                    scope=scope, safe=safe, besteffort=besteffort
                )

            except Exception as ex:

                self._error = ex
                if error:
                    msg = 'Impossible to parse value ({0}) with {1}.'
                    msg = msg.format(self._svalue, self.parser)
                    reraise(Parameter.Error, Parameter.Error(msg))

        return result

    @property
    def value(self):
        """Get parameter value.

        If this cached value is None and this serialized value is not None,
        calculate the new value from the serialized one.

        :return: parameter value.
        :raises: TypeError if serialized value is not an instance of self ptype
            . ParserError if parsing step raised an error.
        """

        result = self._value

        if result is None and self._svalue is not None:

            try:
                result = self._value = self.resolve()

            except Exception as e:
                reraise(
                    Parameter.Error,
                    Parameter.Error('Call the method "resolve" first.')
                )

        return result

    @value.setter
    def value(self, value):
        """Change of parameter value.

        If an error occured, it is stored in this error attribute.

        :param value: new value to use. If input value is not an instance of
            self.ptype, self error
        :raises: TypeError if input value is not an instance of self ptype.
        """

        if value is None or (
                self.ptype is None or isinstance(value, self.ptype)
        ):
            self._value = value

        else:
            # raise wrong type error
            error = TypeError(
                'Wrong value type of {0} ({1}). {2} expected.'.format(
                    self.name, value, self.ptype
                )
            )
            self._error = error
            raise error



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2014 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

__all__ = ['Logger']


from logging import Formatter, getLogger, FileHandler, Filter, INFO

from os import environ
from os.path import join, sep

from b3j0f.utils.property import addproperties

from ..model.cat import Category
from ..model.param import Parameter

from .core import Configurable

B3J0F_LOG_PATH = 'B3J0F_LOG_PATH'  #: log path environment variable.

LOG_PATH = '.'  #: default log path.

# update log_path with environment variable b3j0f log path
if B3J0F_LOG_PATH in environ:
    LOG_PATH = environ[B3J0F_LOG_PATH]


def _filehandler(configurable):
    """Default logging file handler."""

    filename = configurable.log_name.replace('.', sep)
    path = join(configurable.log_path, '{0}.log'.format(filename))

    return FileHandler(path, mode='a+')


def _updatelogger(self, *args, **kwargs):
    """Renew self logger."""

    self._logger = self.newlogger()


@addproperties(
    names=[
        'log_debug_format', 'log_info_format', 'log_warning_format',
        'log_error_format', 'log_critical_format', 'log_name', 'log_path',
        'log_handler'
    ], afset=_updatelogger
)
@addproperties(names=['logger'])
class Logger(Configurable):
    """Manage class conf synchronisation with conf resources."""

    LOG = 'LOG'  #: log attribute name.

    LOG_NAME = 'log_name'  #: logger name property name.
    LOG_LVL = 'log_lvl'  #: logging level property name.
    LOG_PATH = 'log_path'  #: logging path property name.
    LOG_DEBUG_FORMAT = 'log_debug_format'  #: debug log format property name.
    LOG_INFO_FORMAT = 'log_info_format'  #: info log format property name.
    LOG_WARNING_FORMAT = 'log_warning_format'  #: warn log format property name.
    LOG_ERROR_FORMAT = 'log_error_format'  #: error log format property name.
    LOG_CRITICAL_FORMAT = 'log_critical_format'  #: crit log format property.

    # log messages format.
    #: debug message format.
    DEBUG_FORMAT = "[%(asctime)s] [%(levelname)s] [%(name)s] \
[%(process)d] [%(thread)d] [%(pathname)s] [%(lineno)d] %(message)s"
    #: info message format.
    INFO_FORMAT = "[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s"
    WARNING_FORMAT = INFO_FORMAT  #: warning message format.
    ERROR_FORMAT = WARNING_FORMAT  #: error message format.
    CRITICAL_FORMAT = ERROR_FORMAT  #: critical message format.

    DEFAULT_LOG_PATH = '.'  #: default log_path value.
    DEFAULT_LOG_LVL = INFO  #: default log_lvl value.

    def __init__(
            self,
            log_lvl=DEFAULT_LOG_LVL, log_name=None, log_path=DEFAULT_LOG_PATH,
            log_handler=_filehandler, log_info_format=INFO_FORMAT,
            log_debug_format=DEBUG_FORMAT, log_warning_format=WARNING_FORMAT,
            log_error_format=ERROR_FORMAT, log_critical_format=CRITICAL_FORMAT,
            *args, **kwargs
    ):
        """
        :param str log_lvl: logging level. Default is INFO.
        :param str log_name: logger name. Default is configurable class lower
            name.
        :param str log_path: logging file path. Default is current directory.
        :param log_handler: function which takes in parameter self configurable
            and returns a logging Handler.
        :param str log_info_format: info logging level format.
        :param str log_debug_format: debug logging level format.
        :param str log_warning_format: warning logging level format.
        :param str log_error_format: error logging level format.
        :param str log_critical_format: critical logging level format.
        """

        super(Logger, self).__init__(*args, **kwargs)

        # init protected attributes
        self._log_lvl = log_lvl
        self._log_path = log_path
        self._log_handler = log_handler
        self._log_name = log_name if log_name else type(self).__name__.lower()
        self._log_debug_format = log_debug_format
        self._log_info_format = log_info_format
        self._log_warning_format = log_warning_format
        self._log_error_format = log_error_format
        self._log_critical_format = log_critical_format

        if self.logger is None:
            self.logger = self.newlogger()

    def newlogger(self):
        """Get a new logger related to self properties."""

        result = getLogger(self.log_name)
        result.setLevel(self.log_lvl)

        def sethandler(logger, lvl, _format):
            """Set right handler related to input lvl, path and format.

            :param Logger logger: logger on which add an handler.
            :param str lvl: logging level.
            :param str path: file path.
            :param str _format: logging message format.
            """

            class _Filter(Filter):
                """Ensure message will be given for specific lvl."""
                def filter(self, record):
                    return record.levelname == lvl

            # get the rights formatter and filter to set on a file handler
            handler = self._log_handler(self)
            handler.addFilter(_Filter())
            handler.setLevel(lvl)
            formatter = Formatter(_format)
            handler.setFormatter(formatter)

            # if an old handler exist, remove it from logger
            if hasattr(logger, lvl):
                old_handler = getattr(logger, lvl)
                logger.removeHandler(old_handler)

            logger.addHandler(handler)
            setattr(logger, lvl, handler)

        sethandler(result, 'DEBUG', self.log_debug_format)
        sethandler(result, 'INFO', self.log_info_format)
        sethandler(result, 'WARNING', self.log_warning_format)
        sethandler(result, 'ERROR', self.log_error_format)
        sethandler(result, 'CRITICAL', self.log_critical_format)

        return result

    @classmethod
    def clsconf(cls):

        result = Configurable.clsconf()

        result += Category(
            name=Logger.LOG,
            melts=[
                Parameter(name=Logger.LOG_NAME),
                Parameter(name=Logger.LOG_PATH),
                Parameter(name=Logger.LOG_LVL),
                Parameter(name=Logger.LOG_DEBUG_FORMAT),
                Parameter(name=Logger.LOG_INFO_FORMAT),
                Parameter(name=Logger.LOG_WARNING_FORMAT),
                Parameter(name=Logger.LOG_ERROR_FORMAT),
                Parameter(name=Logger.LOG_CRITICAL_FORMAT)
            ]
        )

        return result

    @property
    def log_lvl(self):
        """Get this logger lvl.

        :return: self logger lvl.
        :rtype: str
        """

        return self._log_lvl

    @log_lvl.setter
    def log_lvl(self, value):
        """Change of logging level.

        :param str value: new log_lvl to set up.
        """

        self._log_lvl = value

        self.logger.setLevel(self._log_lvl)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2015 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Parameter parser module.

The parser transforms a serializated parameter value to a parameter value,
related to parameter properties (ptype, scope, safe, etc.).

It takes in parameter:

.. csv-table:
    :header: name, description

    svalue, the serialized value
    conf, the configuration
    configurable, the configurable
    logger, a logger
    scope, dictionary of variables for expression evaluation
    ptype, final value type to check

There are two kind of values:

- simple values: string value like in most system configuration files.
- expression values: contains a dedicated programming language expression
    (python by default).

Simple value
------------

A simple value is a string. It is possible to format a dedicated programming
language expression in respecting this syntax:

*%[{lang}:]{expr}%* where:

- lang is an optional expression evaluator name (``py`` for python, ``js`` for
    javascript, etc.) Default is python.
- expr is the expression in the dedicated programming language.

.. remark::

    In order to use the character ``%`` in the expression, you just have to
    precede it by the *\* character.

Examples:

.. csv-table::

    :header: expression, result, description

    "te%'test'[:-2]%", test, 'te' + 'st' in default (python) evaluator
    "te%js:'test'.substr(2)%", test, 'te' + 'st' in javascript evaluator
    "test number %2\%2%", "test number 0", "default evaluation of 2 modulo 2"

Expression value
----------------

An expression value is an expression given in a dedicated programming language.

Therefore, the result type is not necessary a string. It depends on the
evaluated expression.

Such expression must respects this syntax:

*=[{lang}]:{expr}* where:

- expr is the expression to evaluate.
- lang is the dedicated programming language keyword (for example, ``py`` for
    python, ``js`` for javascript). Default is python.

.. remark::

    in order to use a simple expression starting with *=:*, you juste have to
    add the prefix *\*.

For example:

.. csv-table::
    :header: expression, value


    ``="example"``: string equals "example" from a python interpretor.
    ``=js:"2"``: string equals "2" from javascript interpretor.
    ``=None``: None.
    `=js:3*4``: 12 in javascript.

Reference to other parameters
-----------------------------

It is possible to reference other parameters defined in both configuration files
and configuration.

The syntax is the following:

*'@'[{confpath}/][{cat}\.]{history}{param}* where:

.. csv-table:
    :header: name, description

    confpath, "optional configuration path from where get the parameter. Default
        is current configuration path"
    cat, "optional category name. Default is current category"
    param, "referenced parameter name"
    history, "designates an old parameter from the specified/current category.
        For example:
            - '.' designates the current/specified category.
            - '..' designates the previous one parameter."
            - '...' designates the previous previous one parameter.
            - '....' etc.

For examples:

.. csv-table::
    :header: expr, description

    "@r/c.p", "parameter `p`, from the category `c` in the configuration
        resource `r` (history equals 0 from `c`)"
    "@c.p", "parameter `p` from the category `c` in the same configurable
        scope (history equals 0 from `c`)"
    "@.p", "parameter `p` from the current category. (history equals 0 from
        current category)"
    "@p", "last parameter `p` of the configurable object (history equals 0 from
        last defined category where `p` exists)"
    "@..p", "parameter `p` defined before the current category. Used frequently
        to enrich old parameter value like ``=:@..p + ['new', 'value']``"
    "@c..p", parameter `p`, defined two level before the category `c`.
"""

from __future__ import absolute_import

__all__ = ['parse', 'serialize']


from random import random

from re import compile as re_compile

from six import string_types

from collections import Iterable

from .resolver.core import DEFAULT_BESTEFFORT, DEFAULT_SAFE, DEFAULT_SCOPE

from .resolver.registry import resolve

from parser import ParserError

from json import loads


#: _ref parameter.
EVAL_REF = r'@((?P<path>([^@]|\\@)+)\/)?((?P<cname>\w+)\.)?(?P<history>\.*)(?P<pname>\w+)'

REGEX_REF = re_compile(EVAL_REF)

#: programmatic language expression.
EVAL_FORMAT = r'\%((?P<lang>\w+):)?(?P<expr>([^%]|(\\%))*[^\\])\%'

REGEX_FORMAT = re_compile(EVAL_FORMAT)

EVAL_ANTISLASH = r'(\\(?P<antislash>[@\%\\]))'

EVAL_STR = r'({0})|({1}|({2}))'.format(EVAL_ANTISLASH, EVAL_REF, EVAL_FORMAT)

REGEX_STR = re_compile(EVAL_STR)

EVAL_EXPR = r'=((?P<lang>\w+):)?(?P<expr>.*)$'  #: interpreted expression prefix

REGEX_EXPR = re_compile(EVAL_EXPR)  #: interpreted expression regex comp

EVAL_EXPR_R = r'(({0})|({1}))'.format(EVAL_ANTISLASH, EVAL_REF)

REGEX_EXPR_R = re_compile(EVAL_EXPR_R)


def serialize(expr):
    """Serialize input expr into a parsable value.

    :rtype: str"""

    result = None

    if isinstance(expr, string_types):
        result = expr

    elif expr is not None:
        result = '=py:{0}'.format(expr)

    return result


def parse(
        svalue, conf=None, configurable=None, ptype=None,
        scope=DEFAULT_SCOPE, safe=DEFAULT_SAFE, besteffort=DEFAULT_BESTEFFORT
):
    """Parser which delegates parsing to expression or format parser."""

    result = None

    if ptype is None:
        ptype = object

    compilation = REGEX_EXPR.match(svalue)

    _scope = {} if scope is None else scope.copy()

    if compilation:

        lang, expr = compilation.group('lang', 'expr')

        result = _exprparser(
            expr=expr, lang=lang, conf=conf, configurable=configurable,
            scope=_scope, safe=safe, besteffort=besteffort
        )

    else:
        result = _strparser(
            svalue=svalue, conf=conf, configurable=configurable,
            scope=_scope, safe=safe, besteffort=besteffort
        )

    # try to cast value in ptype
    if not isinstance(result, ptype):
        try:
            result = ptype(result)

        except TypeError:
            result = result

    return result


def _exprparser(
        expr, scope, lang=None, conf=None, configurable=None,
        safe=DEFAULT_SAFE, besteffort=DEFAULT_BESTEFFORT, tostr=False
):
    """In charge of parsing an expression and return a python object."""

    if scope is None:
        scope = {}

    scope.update({
        'configurable': configurable,
        'conf': conf
    })

    expr = REGEX_EXPR_R.sub(
        _refrepl(
            configurable=configurable, conf=conf, safe=safe, scope=scope,
            besteffort=besteffort
        ), expr
    )

    result = resolve(
        expr=expr, name=lang, safe=safe, scope=scope, tostr=tostr,
        besteffort=besteffort
    )

    return result


def _strparser(
        svalue, safe=DEFAULT_SAFE, ptype=None, scope=DEFAULT_SCOPE,
        configurable=None, conf=None, besteffort=DEFAULT_BESTEFFORT
):

    result = REGEX_STR.sub(
        _strrepl(
            safe=safe, scope=scope, configurable=configurable,
            conf=conf, besteffort=besteffort
        ), svalue
    )

    if ptype is None:
        ptype = str

    if not issubclass(ptype, string_types):

        if issubclass(ptype, bool):
            result = result in ('1', 'True', 'true')

        elif issubclass(ptype, dict):
            result = loads(result)

        elif issubclass(ptype, Iterable):
            if result:
                result = ptype((item.strip() for item in result.split(',')))

            else:
                result = ptype()

    return result


def _refrepl(configurable, conf, scope, safe, besteffort):

    def __repl(match, scope=scope, safe=safe, besteffort=besteffort):

        antislash, path, cname, history, pname = match.group(
            'antislash', 'path', 'cname', 'history', 'pname'
        )

        if antislash:
            result = antislash

        else:
            history = (len(history) - 1) if history else 0

            param = _ref(
                configurable=configurable, conf=conf,
                path=path, cname=cname, history=history, pname=pname
            )

            result = '_____{0}'.format(random()).replace('.', '_')

            value = param.resolve(
                configurable=configurable, conf=conf, scope=scope, safe=safe,
                besteffort=besteffort
            )

            scope[result] = value

        return result

    return __repl


def _strrepl(
        configurable, conf,
        scope=DEFAULT_SCOPE, safe=DEFAULT_SAFE, besteffort=DEFAULT_BESTEFFORT
):

    def __repl(match, scope=scope):

        antislash, lang, expr, path, cname, history, pname = match.group(
            'antislash', 'lang', 'expr', 'path', 'cname', 'history', 'pname'
        )

        if antislash:
            result = antislash

        elif expr:
            result = _exprparser(
                expr=expr, lang=lang, conf=conf, configurable=configurable,
                scope=scope, safe=safe, besteffort=besteffort, tostr=True
            )

        else:
            history = (len(history) - 1) if history else 0

            param = _ref(
                configurable=configurable, conf=conf,
                path=path, cname=cname, history=history, pname=pname
            )

            result = param.svalue

        return result

    return __repl


def _ref(
        pname, conf=None, configurable=None, cname=None, path=None, history=0
):
    """Resolve a parameter value.

    :param Configuration conf: configuration to use.
    :param str pname: parameter name.
    :param Configurable configurable: configurable.
    :param str cname: category name.
    :param str path: conf path.
    :param int history: parameter history research.
    :return: parameter.
    :raises: ParserError if conf and configurable are None.
    """

    result = None

    if configurable is not None:

        kwargs = {}

        if conf is not None:
            kwargs['conf'] = conf

        if path is not None:
            kwargs['paths'] = path

        if conf is None:
            conf = configurable.getconf(**kwargs)

    if conf is None:
        raise ParserError(
            'Wrong ref parameters. Conf and configurable are both None.'
        )

    result = conf.param(pname=pname, cname=cname, history=history)

    return result
